@@ -1,7 +1,7 @@
 import asyncio
 import collections
+import contextlib
 import ipaddress
-import os
 import secrets
 import socket
 import struct
@@ -13,6 +13,10 @@ TYPES = collections.namedtuple('Types', [
     'A', 'CNAME', 'TXT', 'AAAA'
 ])(A=1, CNAME=5, TXT=16, AAAA=28)
 
+STRUCT_HEADER = struct.Struct('!HHHHHH')
+STRUCT_TTL_RDATALEN = struct.Struct('!LH')
+STRUCT_QTYPE_QCLASS = struct.Struct('!HH')
+
 # Field names chosen to be consistent with RFC 1035
 Message = collections.namedtuple('Message', [
     'qid', 'qr', 'opcode', 'aa', 'tc', 'rd', 'ra', 'z', 'rcode',
@@ -27,71 +31,60 @@ ResourceRecord = collections.namedtuple('Record', [
     'name', 'qtype', 'qclass', 'ttl', 'rdata',
 ])
 
+
 class ResolverError(Exception):
     pass
 
-class TemporaryResolverError(Exception):
+
+class TemporaryResolverError(ResolverError):
     pass
 
+
 class DoesNotExist(ResolverError):
     pass
 
+
+class PointerLoop(ResolverError):
+    pass
+
+
+class CnameChainTooLong(ResolverError):
+    pass
+
+
 class IPv4AddressTTL(ipaddress.IPv4Address):
     def __init__(self, rdata, expires_at):
         super().__init__(rdata)
-        self._expires_at = expires_at
+        self.expires_at = expires_at
 
-    def ttl(self, now):
-        return max(0.0, self._expires_at - now)
 
 class IPv6AddressTTL(ipaddress.IPv6Address):
     def __init__(self, rdata, expires_at):
         super().__init__(rdata)
-        self._expires_at = expires_at
+        self.expires_at = expires_at
 
-    def ttl(self, now):
-        return max(0.0, self._expires_at - now)
 
 class BytesTTL(bytes):
     def __new__(cls, rdata, expires_at):
         _rdata = super().__new__(cls, rdata)
-        _rdata._expires_at = expires_at
+        _rdata.expires_at = expires_at
         return _rdata
 
-    def ttl(self, now):
-        return max(0.0, self._expires_at - now)
-
-def rdata_ttl(record, ttl_start, min_expires_at):
-    expires_at = min(ttl_start + record.ttl, min_expires_at)
-    return \
-        IPv4AddressTTL(record.rdata, expires_at) if record.qtype == TYPES.A else \
-        IPv6AddressTTL(record.rdata, expires_at) if record.qtype == TYPES.AAAA else \
-        BytesTTL(record.rdata, expires_at)
-
 
 def pack(message):
-    struct_pack = struct.pack
-
-    def pack_string(string):
-        length = len(string)
-        return struct_pack('B%ds' % (length), length, string)
-
     def pack_name(name):
-        return b''.join([
-            pack_string(part)
+        return b''.join(
+            bytes((len(part),)) + part
             for part in name.split(b'.')
-        ]) + b'\0'
+        ) + b'\0'
 
     def pack_resource(record):
         rdata = \
-            b'.'.join(pack_name(record.rdata)) if record.qtype == TYPES.CNAME else \
+            pack_name(record.rdata) if record.qtype == TYPES.CNAME else \
             record.rdata
-        ttl = struct_pack('!L', record.ttl)
-        dl = struct_pack('!H', len(rdata))
-        return ttl + dl + rdata
+        return STRUCT_TTL_RDATALEN.pack(record.ttl, len(rdata)) + rdata
 
-    header = struct_pack(
-        '!HHHHHH',
+    header = STRUCT_HEADER.pack(
         message.qid,
         (message.qr << 15) + (message.opcode << 11) + (message.aa << 10) + (message.tc << 9) +
         (message.rd << 8) + (message.ra << 7) + (message.z << 4) + message.rcode,
@@ -100,21 +93,18 @@ def pack(message):
         len(message.ns),
         len(message.ar),
     )
-    records = b''.join([
-        pack_name(rec.name) + struct_pack('!HH', rec.qtype, rec.qclass)
+    records = b''.join(tuple(
+        pack_name(rec.name) + STRUCT_QTYPE_QCLASS.pack(rec.qtype, rec.qclass)
         for rec in message.qd
-    ] + [
-        pack_name(rec.name) + struct_pack('!HH', rec.qtype, rec.qclass) + pack_resource(rec)
+    ) + tuple(
+        pack_name(rec.name) + STRUCT_QTYPE_QCLASS.pack(rec.qtype, rec.qclass) + pack_resource(rec)
         for group in (message.an, message.ns, message.ar)
         for rec in group
-    ])
+    ))
     return header + records
 
 
 def parse(data):
-    struct_calcsize = struct.calcsize
-    struct_unpack = struct.unpack
-
     def byte(offset):
         return data[offset:offset + 1][0]
 
@@ -123,23 +113,23 @@ def parse(data):
         return offset + length + 1, data[offset + 1:offset + 1 + length]
 
     def load_labels():
-        nonlocal l
+        nonlocal c
 
         followed_pointers = []
-        local_cursor = l
+        local_cursor = c
 
         while True:
-            if byte(local_cursor) >= 192:  # is pointer
+            while byte(local_cursor) >= 192:  # is pointer
                 local_cursor = (byte(local_cursor) - 192) * 256 + byte(local_cursor + 1)
+                if local_cursor in followed_pointers:
+                    raise PointerLoop()
                 followed_pointers.append(local_cursor)
-                if len(followed_pointers) != len(set(followed_pointers)):
-                    raise Exception('Pointer loop')
                 if len(followed_pointers) == 1:
-                    l += 2
+                    c += 2
 
             local_cursor, label = load_label(local_cursor)
             if not followed_pointers:
-                l = local_cursor
+                c = local_cursor
 
             if label:
                 yield label
@@ -152,33 +142,32 @@ def parse(data):
             yield num - (high << length)
             num = high
 
-    def unpack(struct_format):
-        nonlocal l
-        dl = struct_calcsize(struct_format)
-        unpacked = struct_unpack(struct_format, data[l: l + dl])
-        l += dl
+    def unpack(struct_obj):
+        nonlocal c
+        unpacked = struct_obj.unpack_from(data, c)
+        c += struct_obj.size
         return unpacked
 
     def parse_question_record():
         name = b'.'.join(load_labels())
-        qtype, qclass = unpack('!HH')
+        qtype, qclass = unpack(STRUCT_QTYPE_QCLASS)
         return QuestionRecord(name, qtype, qclass)
 
     def parse_resource_record():
-        nonlocal l
+        nonlocal c
         # The start is same as the question record
         name, qtype, qclass = parse_question_record()
-        ttl, dl = unpack('!LH')
+        ttl, dc = unpack(STRUCT_TTL_RDATALEN)
         if qtype == TYPES.CNAME:
             rdata = b'.'.join(load_labels())
         else:
-            rdata = data[l: l + dl]
-            l += dl
+            rdata = data[c: c + dc]
+            c += dc
 
         return ResourceRecord(name, qtype, qclass, ttl, rdata)
 
-    l = 0
-    qid, x, qd_count, an_count, ns_count, ar_count = unpack('!HHHHHH')
+    c = 0
+    qid, x, qd_count, an_count, ns_count, ar_count = unpack(STRUCT_HEADER)
     rcode, z, ra, rd, tc, aa, opcode, qr = split_bits(x, 4, 3, 1, 1, 1, 1, 4, 1)
 
     qd = tuple(parse_question_record() for _ in range(qd_count))
@@ -189,104 +178,65 @@ def parse(data):
     return Message(qid, qr, opcode, aa, tc, rd, ra, z, rcode, qd, an, ns, ar)
 
 
-
-# We implement our own recv/send functions since:
-# - loop.sock_recv doesn't seem to handle cancellation well
-# - There is no asyncio recvfrom/sendto in the standard library, which are
-#   used in tests
-# - We want consistent with the code used in tests
-# - Want to avoid the inflexibility of the streams/protocol/datagram endpoint
-#   framework
-
-async def send_all(loop, sock, data):
-    bytes_sent = await send(loop, sock, data)
-    while bytes_sent != len(data):
-        bytes_sent += await send(loop, sock, data[bytes_sent:])
-
-
-async def send(loop, sock, data):
-    try:
-        return sock.send(data)
-    except BlockingIOError:
-        pass
-
-    fileno = sock.fileno()
-    result = asyncio.Future()
-
-    def write_with_writer():
+async def recvfrom(loop, socks, max_bytes):
+    for sock in socks:
         try:
-            bytes_sent = sock.send(data)
+            return sock.recvfrom(max_bytes)
         except BlockingIOError:
             pass
-        except BaseException as exception:
-            if not result.cancelled():
-                result.set_exception(exception)
-        else:
-            if not result.cancelled():
-                result.set_result(bytes_sent)
 
-    loop.add_witer(fileno, write_with_writer)
+    def reader(sock):
+        def _reader():
+            try:
+                (data, addr) = sock.recvfrom(max_bytes)
+            except BlockingIOError:
+                pass
+            except BaseException as exception:
+                if not result.done():
+                    result.set_exception(exception)
+            else:
+                if not result.done():
+                    result.set_result((data, addr))
+        return _reader
+
+    fileno_socks = tuple((sock.fileno(), sock) for sock in socks)
+    result = asyncio.Future()
+
+    for fileno, sock in fileno_socks:
+        loop.add_reader(fileno, reader(sock))
 
     try:
         return await result
     finally:
-        loop.remove_writer(fileno)
+        for fileno, _ in fileno_socks:
+            loop.remove_reader(fileno)
 
 
-async def recvfrom(loop, sock, max_bytes):
-    # This handles cancellation better than loop.sock_recv, which seems to
-    # causes later sockets on the same fileno to never receive data
-
-    try:
-        return sock.recvfrom(max_bytes)
-    except BlockingIOError:
-        pass
-
-    fileno = sock.fileno()
-    result = asyncio.Future()
-
-    def read_with_reader():
-        try:
-            (data, addr) = sock.recvfrom(max_bytes)
-        except BlockingIOError:
-            pass
-        except BaseException as exception:
-            if not result.cancelled():
-                result.set_exception(exception)
-        else:
-            if not result.cancelled():
-                result.set_result((data, addr))
-
-    loop.add_reader(fileno, read_with_reader)
-
-    try:
-        return await result
-    finally:
-        loop.remove_reader(fileno)
-
-
-def get_nameservers():
+async def get_nameservers_from_etc_resolve_conf(_):
     with open('/etc/resolv.conf', 'r') as file:
-        return tuple(
-            ipaddress.ip_address(words_on_line[1])
-            for words_on_line in [
+        nameservers = tuple(
+            words_on_line[1]
+            for words_on_line in (
                 line.split() for line in file
                 if line[0] not in ['#', ';']
-            ]
+            )
             if len(words_on_line) >= 2 and words_on_line[0] == 'nameserver'
         )
+    for _ in range(5):
+        for nameserver in nameservers:
+            yield (0.5, (nameserver, 53))
 
 
-def get_hosts():
+async def get_host_from_etc_hosts(fqdn, qtype):
     with open('/etc/hosts', 'r') as file:
-        hosts = [
+        hosts = tuple(
             (host, ipaddress.ip_address(words[0]))
             for line in file
-            for (line_before_comment, _, __) in [line.partition('#')]
-            for words in [line_before_comment.split()]
+            for (line_before_comment, _, __) in (line.partition('#'),)
+            for words in (line_before_comment.split(),)
             for host in words[1:]
-        ]
-    return {
+        )
+    hosts = {
         TYPES.A: {
             host.encode(): IPv4AddressTTL(ip_address, expires_at=0)
             for host, ip_address in hosts if isinstance(ip_address, ipaddress.IPv4Address)
@@ -296,9 +246,13 @@ def get_hosts():
             for host, ip_address in hosts if isinstance(ip_address, ipaddress.IPv6Address)
         }
     }
+    try:
+        return hosts[qtype][fqdn]
+    except KeyError:
+        return None
 
 
-def mix_case(fqdn):
+async def mix_case(fqdn):
     return bytes(
         (char | secrets.choice((32, 0))) if 65 <= char < 91 else char
         for char in fqdn.upper()
@@ -306,12 +260,11 @@ def mix_case(fqdn):
 
 
 def Resolver(
-        get_hosts=get_hosts,
-        get_nameservers=get_nameservers,
-        fqdn_transform=mix_case,
-        udp_response_timeout=0.5,
-        udp_attempts_per_server=5,
-    ):
+        get_host=get_host_from_etc_hosts,
+        get_nameservers=get_nameservers_from_etc_resolve_conf,
+        transform_fqdn=mix_case,
+        max_cname_chain_length=20,
+):
 
     loop = \
         asyncio.get_running_loop() if hasattr(asyncio, 'get_running_loop') else \
@@ -323,35 +276,22 @@ def Resolver(
     woken_waiter = {}
 
     async def resolve(fqdn_str, qtype):
-        nameservers = get_nameservers()
-        hosts = get_hosts()
-        fqdn = BytesTTL(fqdn_str.encode(), expires_at=float('inf'))
+        fqdn = BytesTTL(fqdn_str.encode('idna'), expires_at=float('inf'))
 
-        while True:
-            if qtype in hosts and fqdn in hosts[qtype]:
-                return (hosts[qtype][fqdn],)
+        for _ in range(max_cname_chain_length):
+            host = await get_host(fqdn, qtype)
+            if host is not None:
+                return (host,)
 
-            answers = await udp_request_namservers_until_response(nameservers, fqdn, qtype)
-
-            qtype_rdata = tuple(rdata_ttl for rdata_ttl, rdata_qtype in answers if rdata_qtype == qtype)
-            cname_rdata = tuple(rdata_ttl for rdata_ttl, rdata_qtype in answers if rdata_qtype == TYPES.CNAME)
+            cname_rdata, qtype_rdata = await memoized_udp_request(fqdn, qtype)
+            min_expires_at = fqdn.expires_at  # pylint: disable=no-member
             if qtype_rdata:
-                return qtype_rdata
-            elif cname_rdata:
-                fqdn = cname_rdata[0]
-            else:
-                raise DoesNotExist()
+                return rdata_ttl_min_expires(qtype_rdata, min_expires_at)
+            fqdn = rdata_ttl_min_expires([cname_rdata[0]], min_expires_at)[0]
 
-    async def udp_request_namservers_until_response(nameservers, fqdn, qtype):
-        exception = None
-        for addr in nameservers:
-            try:
-                return await memoized_udp_request(addr, fqdn, qtype)
-            except (asyncio.TimeoutError, TemporaryResolverError) as recent_exception:
-                exception = recent_exception
-        raise exception
+        raise CnameChainTooLong()
 
-    async def memoized_udp_request(addr, fqdn, qtype):
+    async def memoized_udp_request(fqdn, qtype):
         """Memoized udp_request, that allows a dynamic expiry for each result
 
         Multiple callers for the same args will wait for first call to
@@ -362,6 +302,13 @@ def Resolver(
         it. A non-cancellation exception is propagated to all callers
         """
 
+        key = (fqdn, qtype)
+
+        try:
+            return cache[key]
+        except KeyError:
+            pass
+
         def wake_next():
             # Find the next non cancelled...
             while waiter_queue and waiter_queue[0].cancelled():
@@ -376,13 +323,6 @@ def Resolver(
                 # Delete the queue only if we haven't woken anything up
                 del waiter_queues[key]
 
-        key = (addr, fqdn, qtype)
-
-        try:
-            return cache[key]
-        except KeyError:
-            pass
-
         if key not in waiter_queues:
             waiter_queue = collections.deque()
             waiter_queues[key] = waiter_queue
@@ -399,13 +339,13 @@ def Resolver(
                 raise
             else:
                 if has_other_task_result:
-                   return other_task_result
+                    return other_task_result
             finally:
                 if key in woken_waiter and waiter == woken_waiter[key]:
                     del woken_waiter[key]
 
         try:
-            answers = await udp_request(addr, fqdn, qtype)
+            answers = await udp_request_namservers_until_response(fqdn, qtype)
 
         except asyncio.CancelledError:
             wake_next()
@@ -428,31 +368,36 @@ def Resolver(
                     waiter.set_result((True, answers))
             del waiter_queues[key]
 
-            expires_at = min(rdata_ttl._expires_at for rdata_ttl, _ in answers)
-            invalidate_callback = loop.call_at(expires_at, invalidate, key)
+            expires_at = min(
+                rdata_ttl.expires_at
+                for rdata_groups in answers
+                for rdata_ttl in rdata_groups
+            )
+            invalidate_callbacks[key] = loop.call_at(expires_at, invalidate, key)
             cache[key] = answers
-            invalidate_callbacks[key] = invalidate_callback
             return answers
 
     def invalidate(key):
         del cache[key]
-        invalidate_callbacks[key].cancel()
-        del invalidate_callbacks[key]
+        invalidate_callbacks.pop(key).cancel()
 
     def invalidate_all():
-        for key, _ in list(cache.items()):
-            invalidate(key)
+        for callback in invalidate_callbacks.values():
+            callback.cancel()
+        invalidate_callbacks.clear()
+        cache.clear()
 
-    async def udp_request(addr, fqdn, qtype):
-        exception = None
-        for _ in range(udp_attempts_per_server):
+    async def udp_request_namservers_until_response(fqdn, qtype):
+        exception = ResolverError()
+        async for nameserver in get_nameservers(fqdn):
+            timeout, addrs = nameserver[0], nameserver[1:]
             try:
-                return await timeout_udp_request_attempt(addr, fqdn, qtype)
+                return await timeout_udp_request_attempt(timeout, addrs, fqdn, qtype)
             except (asyncio.TimeoutError, TemporaryResolverError) as recent_exception:
                 exception = recent_exception
         raise exception
 
-    async def timeout_udp_request_attempt(addr, fqdn, qtype):
+    async def timeout_udp_request_attempt(timeout, addrs, fqdn, qtype):
         cancelling_due_to_timeout = False
         current_task = \
             asyncio.current_task() if hasattr(asyncio, 'current_task') else \
@@ -463,65 +408,103 @@ def Resolver(
             cancelling_due_to_timeout = True
             current_task.cancel()
 
-        handle = loop.call_later(udp_response_timeout, cancel)
+        handle = loop.call_later(timeout, cancel)
 
         try:
-            return await udp_request_attempt(addr, fqdn, qtype)
+            return await udp_request_attempt(addrs, fqdn, qtype)
         except asyncio.CancelledError:
             if cancelling_due_to_timeout:
                 raise asyncio.TimeoutError()
-            else:
-                raise
-                
+            raise
+
         finally:
             handle.cancel()
 
-    async def udp_request_attempt(addr, fqdn, qtype):
-        qid = secrets.randbelow(65536)
-        fqdn_transformed = fqdn_transform(fqdn)
-        req = Message(
-            qid=qid, qr=QUESTION, opcode=0, aa=0, tc=0, rd=1, ra=0, z=0, rcode=0,
-            qd=(QuestionRecord(fqdn_transformed, qtype, qclass=1),), an=(), ns=(), ar=(),
-        )
-        packed = pack(req)
+    async def udp_request_attempt(addrs, fqdn, qtype):
+        async def req():
+            qid = secrets.randbelow(65536)
+            fqdn_transformed = await transform_fqdn(fqdn)
+            return Message(
+                qid=qid, qr=QUESTION, opcode=0, aa=0, tc=0, rd=1, ra=0, z=0, rcode=0,
+                qd=(QuestionRecord(fqdn_transformed, qtype, qclass=1),), an=(), ns=(), ar=(),
+            )
+
+        with contextlib.ExitStack() as stack:
+            socks = tuple(
+                stack.enter_context(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
+                for addr in addrs
+            )
+
+            connections = {}
+            last_exception = OSError()
+            for addr_port, sock in zip(addrs, socks):
+                try:
+                    sock.setblocking(False)
+                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 512)
+                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 512)
+                    sock.connect(addr_port)
+                except OSError as exception:
+                    last_exception = exception
+                else:
+                    connections[addr_port] = (sock, await req())
+
+            if not connections:
+                raise last_exception
 
-        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
-            sock.setblocking(False)
-            sock.connect((str(addr), 53))
             ttl_start = loop.time()
-            await send_all(loop, sock, packed)
+            for (sock, req) in connections.values():
+                await loop.sock_sendall(sock, pack(req))
 
-            while True:  # We might be getting spoofed messages
-                response_data, _ = await recvfrom(loop, sock, 512)
+            while connections:
+                connected_socks = tuple(sock for sock, req in connections.values())
+                response_data, addr_port = await recvfrom(loop, connected_socks, 512)
 
-                # Some initial peeking before parsing
-                if len(response_data) < 12:
-                    continue
-                qid_matches = req.qid == struct.unpack('!H', response_data[:2])[0]
-                if not qid_matches:
+                try:
+                    res = parse(response_data)
+                except (struct.error, IndexError, PointerLoop):
                     continue
 
-                res = parse(response_data)
                 trusted = res.qid == req.qid and res.qd == req.qd
-
                 if not trusted:
                     continue
 
+                del connections[addr_port]
+
                 name_error = res.rcode == 3
                 non_name_error = res.rcode and not name_error
-                answers = [
-                    (rdata_ttl(answer, ttl_start, fqdn._expires_at), answer.qtype)
+                name_lower = req.qd[0].name.lower()
+                cname_answers = tuple(
+                    rdata_ttl(answer, ttl_start)
                     for answer in res.an
-                    if answer.name == fqdn_transformed
-                ]
-
+                    if answer.name.lower() == name_lower and answer.qtype == TYPES.CNAME
+                )
+                qtype_answers = tuple(
+                    rdata_ttl(answer, ttl_start)
+                    for answer in res.an
+                    if answer.name.lower() == name_lower and answer.qtype == qtype
+                )
                 if non_name_error:
-                    raise TemporaryResolverError()
-                elif name_error or not answers:
+                    continue
+                elif name_error or (not cname_answers and not qtype_answers):
                     # a name error can be returned by some non-authoritative
                     # servers on not-existing, contradicting RFC 1035
                     raise DoesNotExist()
                 else:
-                    return answers
+                    return cname_answers, qtype_answers
+
+            raise TemporaryResolverError()
+
+    def rdata_ttl(record, ttl_start):
+        expires_at = ttl_start + record.ttl
+        return \
+            IPv4AddressTTL(record.rdata, expires_at) if record.qtype == TYPES.A else \
+            IPv6AddressTTL(record.rdata, expires_at) if record.qtype == TYPES.AAAA else \
+            BytesTTL(record.rdata, expires_at)
+
+    def rdata_ttl_min_expires(rdata_ttls, expires_at):
+        return tuple(
+            type(rdata_ttl)(rdata=rdata_ttl, expires_at=min(expires_at, rdata_ttl.expires_at))
+            for rdata_ttl in rdata_ttls
+        )
 
     return resolve, invalidate_all
